from functools import wraps


def validate_sns_message_signature(func):
    """
    This decorator is used to validate NSN message authenticity, auto-confirm SNS subscriptions and unsubscribes to the
    endpoint as well as convert message bodies to json.
    if applicable
    """
    @wraps(func)
    def inner(self, request, params, *args, **kwargs):
        import json
        from sns_message_validator import (
            InvalidCertURLException,
            InvalidSignatureVersionException,
            SignatureVerificationFailureException,
            SNSMessageValidator,
        )
        from rest_framework import exceptions
        if params is None:
            params = self.nor
        sns_message_validator = SNSMessageValidator()
        try:
            sns_message_validator.validate_message(message=params)
        except InvalidCertURLException:
            raise exceptions.ValidationError('Invalid certificate URL.')
        except InvalidSignatureVersionException:
            raise exceptions.ValidationError('Unexpected signature version.')
        except SignatureVerificationFailureException:
            raise exceptions.ValidationError('Failed to verify signature.')

        ''' Confirm subscribe on initial request'''
        if params.get('Type', 'None') in ['SubscriptionConfirmation']:
            import urllib.request
            confirmation_url = params.get('SubscribeURL')
            contents = urllib.request.urlopen(confirmation_url).read()

        ''' Confirm Unsubscribe on request'''
        if params.get('Type', 'None') in ['UnsubscribeConfirmation']:
            import urllib.request
            confirmation_url = params.get('UnsubscribeURL')
            contents = urllib.request.urlopen(confirmation_url).read()

        ''' Attempt parse of message into dict else original string '''
        try:
            json_message = json.loads(params['Message'])
        except Exception as e:
            pass
        else:
            params['Message'] = json_message
        ''' Not a subscription request message '''
        return func(self, request, params, *args, **kwargs)
    return inner


def determine_staff(func):
    """
    Checks the presence of a "token" request parameter and tries to
    authenticate the user based on its content.
    """
    @wraps(func)
    def inner(self, request, *args, **kwargs):

        return func(self, request, *args, **kwargs)
    return inner


from django.contrib import admin
from .models import Topic, TopicCallback, Message


class TopicAdmin(admin.ModelAdmin):
    list_display = ['id', 'topic', 'arn']
    list_display_links = ['topic']


class TopicCallbackAdmin(admin.ModelAdmin):
    list_display = ['id', 'topic', 'callback', 'arn']
    list_display_links = ['topic', ]


class MessageAdmin(admin.ModelAdmin):
    list_display = ['topic', 'message_id', 'source', 'content_type']
    readonly_fields = ['message_id', 'content_type', 'object_id', 'topic', 'subject', 'message', 'source']
    list_filter = ('created', 'content_type')


admin.site.register(Topic, TopicAdmin)
admin.site.register(TopicCallback, TopicCallbackAdmin)
admin.site.register(Message, MessageAdmin)



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models


# Create your models here.


class Topic(models.Model):
    topic = models.CharField(max_length=128, help_text='Topic you are subscribing to.')
    arn = models.CharField(max_length=255, blank=True, null=True, editable=False)

    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        verbose_name = "Topic"
        verbose_name_plural = "Topics"

    def __str__(self):
        return self.topic

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        try:
            from .topic_providers import MessageWrapper
            mw = MessageWrapper.get('sns')
            arn = mw.create_topic(topic=self.topic)
            self.arn = arn.arn
        except Exception as e:
            pass  # todo: log this issue
        super().save(force_insert, force_update, using, update_fields)


class TopicCallback(models.Model):
    TYPES = (
        ('https', 'HTTPS',),
        ('http', 'HTTP',),
        ('email', 'Email',),
        ('lambda', 'AWS Lambda',),
        ('sqs', 'Amazon SQS',),
    )
    topic = models.ForeignKey(Topic, on_delete=models.SET_NULL, related_name='subscriptions', null=True)
    protocol = models.CharField(max_length=32, choices=TYPES)
    callback = models.URLField(max_length=255, help_text='Fully qualified callback URI that messages for the topic will\
     be pushed to')
    enabled = models.BooleanField(default=True)  # future use to disable the subscription
    arn = models.CharField(max_length=255, blank=True, null=True, editable=False)

    created = models.DateTimeField(auto_now_add=True, db_index=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        verbose_name = "Topic Callback"
        verbose_name_plural = "Topic Callbacks"

    def __str__(self):
        return self.topic.__str__()

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        try:
            if self.enabled is True:
                from .topic_providers import MessageWrapper
                mw = MessageWrapper.get('sns')
                arn = mw.subscribe(topic=self.topic.topic, protocol=self.protocol, endpoint=self.callback)
                self.arn = arn.arn
        except Exception as e:
            pass  # TODO: Add some kind of logging to make system aware of this failure
        super().save(force_insert, force_update, using, update_fields)


class Message(models.Model):
    topic = models.ForeignKey(Topic, on_delete=models.SET_NULL, null=True, related_name='sent_messages')
    message = models.JSONField()
    subject = models.CharField(max_length=128, blank=True, null=True)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE, related_name='sent_messages', null=True)
    object_id = models.CharField(max_length=128, null=True)
    source = GenericForeignKey('content_type', 'object_id')
    message_id = models.UUIDField(null=True, )

    created = models.DateTimeField(auto_now_add=True, db_index=True)

    class Meta:
        verbose_name = "SentMessage"
        verbose_name_plural = "SentMessages"

    def __str__(self):
        return self.topic.__str__()

    @classmethod
    def publish_message(cls, topic: str, message, subject: str = None, source=None, *args, **kwargs):
        from .topic_providers import MessageWrapper
        mw = MessageWrapper.get('sns')
        topic, created = Topic.objects.get_or_create(topic=topic)
        _message = cls(topic=topic, message=message, subject=subject, source=source)
        try:
            resp = mw.publish(topic, message, subject=subject, *args, **kwargs)
            _message.message_id = resp['MessageId']
        except Exception as e:
            print(e)
            pass  # todo: add some kind of logging for this failure
        _message.save()



import urllib.request


def check_and_register_sns(request, params):
    """
    Checks if an inbound message is an SNS confirm registration message, and then completes the registration
    :return:
    """
    if params.get('Type', 'None') not in ['SubscriptionConfirmation']:
        ''' Not a subscription request message '''
        return False

    confirmation_url = params.get('SubscribeURL')
    contents = urllib.request.urlopen(confirmation_url).read()
    return True




from abc import ABC, abstractmethod
from typing import Union


class MessageWrapper():
    """ Abstract class to define methods that must be overwritten by children """

    @abstractmethod
    def create_topic(self, topic: str) -> bool:
        return True

    @abstractmethod
    def delete_topic(self, topic: str) -> bool:
        return True

    @abstractmethod
    def subscribe(self, topic: str) -> bool:
        return True

    @abstractmethod
    def unsubscribe(self, topic: str, topic_link: Union[str, bytes, None] = None, *args, **kwargs) -> bool:
        return True

    @abstractmethod
    def publish(self, topic: str, message: Union[str, dict], subject: str = None, *args, **kwargs):
        return ""

    @classmethod
    def get(cls, message_type='sns', **kwargs):

        from importlib import import_module
        try:
            ret_cls = import_module(f'messaging.topic_providers.{message_type}')
            return ret_cls.MessageWrapper(**kwargs)
        except Exception as e:
            return cls()



import logging

from django.http import HttpResponse, StreamingHttpResponse, QueryDict
from rest_framework import permissions, exceptions
from rest_framework.parsers import BaseParser
from rest_framework.views import APIView, Response

from .decorators import validate_sns_message_signature, determine_staff
from .serializers import APISerializer

logger = logging.getLogger('ddyxdebug')


class PlainTextOrJsonParser(BaseParser):
    """
    Plain text parser.
    """
    media_type = 'text/plain'

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Attempt to parse the string as JSON failing which return a string representing the body of the request.
        Written to handle messages from AWS SNS, but may be needed for TCS inbound notifications
        """
        import json
        raw = stream.read()
        try:
            data = json.loads(raw)
        except:
            data = raw.decode('utf-8')

        return data


class ActionAPIView(APIView):
    """
    Base class for action services taken by the new front end.
    The action function will be determined, if it exists and called.
    Each action function takes the request, expecting JSON body (if required) and returns a dictionary
    or Response instance
    """
    permission_classes = [permissions.IsAuthenticated]

    _last_action = None
    success = True
    RESPONSE_KEYS = ['user', 'action', 'token', 'subject', 'type', 'messageid']
    raw_body = ''

    def post(self, request, action, **kwargs):
        return self.get(request, action, **kwargs)

    @determine_staff
    def get(self, request, action, **kwargs):
        params = self.normalize_params(request)
        kwargs['params'] = params
        self._last_action = params.get('action', action)
        try:
            lv_action = self.__getattribute__(self._last_action)
        except AttributeError:
            lv_action = self.action_does_not_exist
        response = lv_action(request, **kwargs)
        if isinstance(response, Response):
            response = response.data
        if isinstance(response, (HttpResponse, StreamingHttpResponse)):
            return response

        response_context = dict(
            [item for item in list({
                k: params.get(k, None) for k in self.RESPONSE_KEYS
            }.items()) if item[1] is not None])
        serialised = APISerializer(response, context=response_context)
        return Response(serialised.data)

    def action_does_not_exist(self, *args, **kwargs):
        return {'success': False, 'reason': "Action '%s' does not exist." % self._last_action}

    def normalize_params(self, request):
        """
        Combines the possible parameter sources from the request.
        :param request: a REST API request object
        :return: a combined dict of the 2 possible data sources from the request
        """
        try:
            self.raw_body = request._request.body
        except:
            self.raw_body = ""
        params = request.query_params.dict().copy()
        if isinstance(request.data, QueryDict):
            params.update(request.data.dict())
        else:
            params.update(request.data)
        return params

    def check_roles(self, valid_roles, section='this task'):
        if self.staff.roles.filter(id__in=valid_roles).count() == 0:
            raise exceptions.PermissionDenied('You do not have sufficient right to perform %s.' % section)


class SNSActionAPIView(ActionAPIView):
    permission_classes = []
    parser_classes = [PlainTextOrJsonParser]

    def get(self, request, action, **kwargs):
        return super().get(request, action, **kwargs)


class SnsMessageActionAPIViewset(SNSActionAPIView):
    ''' Acts as an example of how to implement a self confirming SNS endpoint '''
    @validate_sns_message_signature
    def echo(self, request, params, *args, **kwargs):
        logger.info(params)
        return params

    @validate_sns_message_signature
    def test_registration(self, request, params, *args, **kwargs):
        logger.info(params)

        return params


# Generated by Django 3.2.13 on 2022-07-25 12:15

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('messaging', '0002_add_topic_and_message'),
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='sent_messages', to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='message',
            name='object_id',
            field=models.CharField(max_length=128, null=True),
        ),
    ]
